import random
import time
from typing import Literal, TypeAlias, NoReturn, List
from xingyun.universal.import_module import my_import_module

'''literals that are allowed in the `module` parameter.'''
RandomAllowedModule: TypeAlias = Literal["torch"] | Literal["numpy"] | Literal["random"] 

def set_module_seed(seed: int , module: RandomAllowedModule):
    '''set random seed of a specific module.'''
    
    if module == "torch":
        torch = my_import_module("torch")
        cuda  = my_import_module("torch.cuda")

        if torch is None or cuda is None:
            return 

        torch.manual_seed(seed)
        cuda.manual_seed_all(seed)

        backends = my_import_module("torch.backends")
        backends.cudnn.deterministic = True
        backends.cudnn.benchmark = False
    if module == "numpy":
        np = my_import_module("numpy")

        if np is None:
            return 
        
        np.random.seed(seed)
        
    if module == "random":
        random.seed(seed)


def set_random_seed(seed: int , modules: List[RandomAllowedModule] = ["random" , "torch" , "numpy"]):
    '''Set random seed.
        
    ### Parameters
        - seed: random seed.
        - modules: Which module to set. Support `random`, `torch` and `numpy`.
            Notice that the random seed of `torch.cuda` will also be set.
    '''
    [set_module_seed(seed, m) for m in modules]



import time
import copy
import random
from typing import Any
from xingyun.universal.import_module import my_import_module

from .set_random_seed import RandomAllowedModule

def get_random_state(module: RandomAllowedModule) -> Any:
    if module == "random":
        return random.getstate()
    
    if module == "numpy":
        np = my_import_module("numpy")
        return np.random.get_state()
    
    if module == "torch":
        torch = my_import_module("torch")
        cuda = my_import_module("torch.cuda")
        
        return {
            "torch": torch.random.get_rng_state(),
            "cuda" : cuda .random.get_rng_state(),
        }

def set_random_state(state: Any, module: RandomAllowedModule) -> bool:
    flag = True
    if module == "random":
        try:
            random.setstate(state)
        except:
            flag = False
    
    if module == "numpy":
        try:
            np = my_import_module("numpy")
            np.random.set_state(state)
        except:
            flag = False
    
    if module == "torch":
        try:
            torch = my_import_module("torch")
            cuda = my_import_module("torch.cuda")

            torch.random.set_rng_state(state["torch"])
            cuda.random.set_rng_state(state["cuda"])
            
        except:
            flag = False
    
    if not flag:
        raise RuntimeError(f"set random state of module {module} bad.")

    return flag

class MyRandom:
    def __init__(self, random_seed: int | None = None, modules:  list[RandomAllowedModule] = ["random" , "torch" , "numpy"]):
        '''This class create a temporary environment, inside which the random seed 
        is set to a given value while not affecting the global random seed.

        Notice that, to make this class work, the global random seed must be also managed by `xingyun`.
        '''
        if random_seed is None:
            random_seed = int( time.time() )
    
        self.random_seed = random_seed
        self.modules = modules

        self.entering_state = {}

    def __enter__(self):
        for m in self.modules:
            self.entering_state[m] = get_random_state(m)
    
    def __exit__(self, *args, **kwargs):
        for m in self.modules:
            set_random_state(self.entering_state[m],m)
            


        






from xingyun.time.my_clock import my_clock
import time
from typing import Callable, List, Any

class Logger:
	'''This is a class to make easy log.'''

	def __init__(self , 
			targets: List[Callable[[str], Any]] = [print] , 
			preprocesses: List[Callable[[str], str]] = [
				lambda s: ("[ %.2fs ] " % my_clock()) + s
			], 
		):
		'''The intialization function class `Logger`.
	
        ### Parameters
            - targets: List of callable objects. Each `log()` will call each target once.
            - prefix: will be called and add to the start of each line. Default to current time.
            - postfix: will be called and add to the end of each line.	
        '''

		self.targets  	 = targets
		self.preprocesses  = preprocesses

	def log(self , content: Any = ""):
		content = str(content)

		content = self.apply_preprocess(content)
		for tar in self.targets:
			tar(content)

	def log_separator(self , num: int = 50 , char: str = "-"):
		'''out put `num` times `char`.
		'''
		self.log(char * num)

	def apply_preprocess(self , content):

		for p in self.preprocesses:
			content = p(content)

		return content
	



'''
This module helps easy use AWS S3.
'''

import boto3
import tempfile
import pickle
from typing import Literal, Any
import blosc

class AWSS3:
    def __init__(self, key_id: str, key_sec: str, region: str, bucket: str):
        self.s3 = boto3.client("s3" , 
            region_name             = region ,
            aws_access_key_id       = key_id , 
            aws_secret_access_key   = key_sec , 
        )
        self.bucket = bucket

    def __del__(self):
        self.s3.close()

    def set(self, data: Any, tar_path: str, format: Literal["binary" , "str" , "pickle"] = "pickle"):
        '''Set the content of a remote file.

        ### Returns
            Return `True` if success, else return `False`.
        '''
        tar_path = str(tar_path).replace("\\"  , "/")

        if format == "str":
            try:
                data = bytes(str(data), encoding = "utf-8")
                with tempfile.TemporaryFile(mode = "wb+") as fil:
                    fil.write(data)
                    fil.seek(0,0) # 文件指针移动到开头
                    self.s3.upload_fileobj(fil , self.bucket , tar_path)
            except Exception:
                return False
        if format == "pickle":
            data = pickle.dumps(data)
            data = blosc.compress(data) # compress data

        if format == "pickle" or format == "pickle":
            try:
                with tempfile.TemporaryFile(mode = "wb+") as fil:
                    fil.write(data)
                    fil.seek(0,0) # 文件指针移动到开头
                    self.s3.upload_fileobj(fil , self.bucket , tar_path)
            except Exception:
                return False

        return True

    def get(self, tar_path: str, format: Literal["binary" , "str" , "pickle"] = "pickle") -> Any:
        '''Get the content of a remote file. 
        
        ### Returns
            Return `None` if fail.
        '''

        tar_path = str(tar_path).replace("\\"  , "/")
        data = None
        if format == "str":
            try:
                with tempfile.TemporaryFile(mode = "wb+") as fil:
                    self.s3.download_fileobj(self.bucket , tar_path, fil)
                    fil.seek(0,0) # 文件指针移动到开头
                    data = fil.read()
                    data = str(data, encoding = "utf-8")
            except Exception:
                return None
        if format == "pickle" or format == "bianry":
            try:
                with tempfile.TemporaryFile(mode = "wb+") as fil:
                    self.s3.download_fileobj(self.bucket , tar_path, fil)
                    fil.seek(0,0) # 文件指针移动到开头
                    data = fil.read()
                    if format == "pickle":
                        try:
                            data = blosc.decompress(data)
                        except: 
                            pass # not blosc compressed. do nothing
                        data = pickle.loads(data)
            except Exception:
                return None
        return data




def convert_size_to_bytes(size_str: str) -> int:
    '''convert a string that describe size to number of bytes.'''
    units = {"b": 1, "kb": 1024, "mb": 1024**2, "gb": 1024**3, "tb": 1024**4}
    size_str = size_str.strip().lower()

    num  = "".join(filter(lambda x: x.isdigit(), size_str.lower()))
    unit = "".join(filter(lambda x: x.isalpha(), size_str.lower()))

    if unit not in units:
        raise RuntimeError(f"Invalid unit {unit}. Use 'b', 'kb', 'mb', 'gb', or 'tb'.")

    return int(float(num) * units[unit])
